#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module contains tools for metrics transformations during a
preprocessing task.
"""
from typing import Dict, Union

import numpy as np
import pandas as pd
import scipy.stats as sps

from ambrosia import types
from ambrosia.tools.ab_abstract_component import AbstractFittableTransformer
from ambrosia.tools.back_tools import wrap_cols


class BoxCoxTransformer(AbstractFittableTransformer):
    """
    Unit for a Box-Cox transformation of the pandas data.

    A Box Cox transformation helps to transform non-normal dependent variables
    into a normal shape. All variables values must be positive.

    Optimal transformation lambdas are selected automatically during
    the transformer fit process.


    Attributes
    ----------
    column_names : List
        Names of column which will be selected for the transformation.
    lambda_ : np.ndarray
        Array of parameters using during the transformation of the
        selected columns.
    fitted : bool
        Fit flag.

    Examples
    --------
    >>> boxcox = BoxCoxTransformer()
    >>> boxcox.fit(dataframe, ['column1', 'column2'])
    >>> boxcox.transform(dataframe, inplace=True)

    """

    def __str__(self) -> str:
        return "Box-Cox transformation"

    def __init__(
        self,
    ) -> None:
        """
        BoxCoxTransformer class constructor.
        """
        self.column_names = None
        self.lambda_ = None
        super().__init__()

    def __calculate_lambda_(
        self,
        dataframe: pd.DataFrame,
    ) -> None:
        columns_num: int = len(self.column_names)
        self.lambda_ = np.zeros(columns_num)
        X: np.ndarray = dataframe[self.column_names].values
        for num in range(columns_num):
            self.lambda_[num] = sps.boxcox(X[:, num])[1]

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            "column_names": self.column_names,
            "lambda_": self.lambda_.tolist(),
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load instance parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """
        if "column_names" in params:
            self.column_names = params["column_names"]
        else:
            raise TypeError(f"params argument must contain: {'column_names'}")
        if "lambda_" in params:
            self.lambda_ = np.array(params["lambda_"])
        else:
            raise TypeError(f"params argument must contain: {'lambda_'}")
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
    ):
        """
        Fit to calculate transformation parameters for the selected columns.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to calculate optimal transformation parameters.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.

        Returns
        -------
        self : object
            Instance object.
        """
        self.column_names = wrap_cols(column_names)
        self._check_cols(dataframe, self.column_names)
        self.__calculate_lambda_(dataframe)
        self.fitted = True
        return self

    def transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Apply Box-Cox transformation for the data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.column_names)
        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        X: np.ndarray = transformed[self.column_names].values
        for num in range(len(self.column_names)):
            if self.lambda_[num] == 0:
                X[:, num] = np.log(X[:, num])
            else:
                X[:, num] = (X[:, num] ** self.lambda_[num] - 1) / self.lambda_[num]
        transformed[self.column_names] = X
        return None if inplace else transformed

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Fit transformer parameters using given dataframe and transform it.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe for calculation of optimal parameters and further
            transformation.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self.fit(dataframe, column_names)
        return self.transform(dataframe, inplace)

    def inverse_transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Apply inverse Box-Cox transformation for the data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to inverse transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.column_names)
        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        X_tr: np.ndarray = transformed[self.column_names].values
        for num in range(len(self.column_names)):
            if self.lambda_[num] == 0:
                X_tr[:, num] = np.exp(X_tr[:, num])
            else:
                X_tr[:, num] = (X_tr[:, num] * self.lambda_[num] + 1) ** (1 / self.lambda_[num])
        transformed[self.column_names] = X_tr
        return None if inplace else transformed


class LogTransformer(AbstractFittableTransformer):
    """
    Unit for a logarithmic transformation of the pandas data.

    A logarithmic transformation helps to transform some metrics distributions
    into a more normal shape and reduce the variance.
    All metrics values must be positive.


    Attributes
    ----------
    column_names : List
        Names of column which will be selected for the transformation.
    fitted : bool
        Fit flag.

    Examples
    --------
    >>> log = LogTransformer()
    >>> log.fit(dataframe, ['column1', 'column2'])
    >>> log.transform(dataframe, inplace=True)

    """

    def __str__(self) -> str:
        return "Logarithmic transformation"

    def __init__(self) -> None:
        """
        LogTransformer class constructor.
        """
        self.column_names = None
        super().__init__()

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.
        """
        self._check_fitted()
        return {
            "column_names": self.column_names,
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load instance parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """
        if "column_names" in params:
            self.column_names = params["column_names"]
        else:
            raise TypeError(f"params argument must contain: {'column_names'}")
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
    ):
        """
        Fit names of the selected columns.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe with metrics.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.

        Returns
        -------
        self : object
            Instance object.
        """
        self.column_names = wrap_cols(column_names)
        self._check_cols(dataframe, self.column_names)
        self.fitted = True
        return self

    def transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Apply log transformation for the data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.column_names)
        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        if (transformed[self.column_names] > 0).all(axis=None):
            transformed[self.column_names] = np.log(transformed[self.column_names].values)
        else:
            raise ValueError(f"All values in columns {self.column_names} must be positive")
        return None if inplace else transformed

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        column_names: types.ColumnNamesType,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Fit transformer parameters using given dataframe and transform it.

        Only column names are fittable.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to transform.
        column_names : ColumnNamesType
            One or number of columns in the dataframe.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self.fit(dataframe, column_names)
        return self.transform(dataframe, inplace)

    def inverse_transform(self, dataframe: pd.DataFrame, inplace: bool = False) -> Union[pd.DataFrame, None]:
        """
        Apply inverse log transformation for the data.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Dataframe to inverse transform.
        inplace : bool, default: ``False``
            If ``True`` transforms the given dataframe, otherwise copy and
            returns an another one.

        Returns
        -------
        df : Union[pd.DataFrame, None]
            Transformed dataframe or None
        """
        self._check_fitted()
        self._check_cols(dataframe, self.column_names)
        transformed: pd.DataFrame = dataframe if inplace else dataframe.copy()
        transformed[self.column_names] = np.exp(transformed[self.column_names].values)
        return None if inplace else transformed



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Callable, Iterable, List, Tuple, Union

import numpy as np
import scipy.stats as sps

from ambrosia import types

ADMISSIBLE_TRANSFORMATIONS: List[str] = ["fraction"]


def calculate_point_effect_by_delta_method(
    mean_a: float,
    mean_b: float,
    var_group_a: float,
    var_group_b: float,  # pylint: disable=W0613
    covariance_ab: float,
    mean_size: int,  # Same size for both groups, TODO to change?
    transformation: str,
) -> float:
    """
    Calculate pvalue for continuous transformation using Delta method.

    Applying the Delta Method in Metric Analytics: A Practical Guide with Novel Ideas.
    Alex Deng, Ulf Knoblich, Jiannan Lu. 2018
    """
    if transformation == "fraction":
        fraction_estimation: float = mean_b / mean_a
        bias_correction: float = (mean_b / mean_a**3) * (var_group_a / mean_size) - (1.0 / mean_a**2) * (
            covariance_ab / mean_size
        )
        point_estimate = fraction_estimation - 1 + bias_correction
    else:
        raise ValueError(f"Got unknown random variable transformation: {ADMISSIBLE_TRANSFORMATIONS}")
    return point_estimate


def calc_statistic_for_delta_method(
    mean_a: float, mean_b: float, var_group_a: float, var_group_b: float, covariance_ab: float, size: int
) -> float:
    """
    Helps calculate statistic after delta method transformation
    """
    return np.sqrt(
        np.abs(
            (var_group_b / (mean_a**2) - 2 * covariance_ab * mean_b / mean_a**3 + var_group_a * (mean_b**2) / mean_a**4)
            / size
        )
    )


def calculate_pvalue_by_delta_method(
    mean_a: float,
    mean_b: float,
    var_group_a: float,
    var_group_b: float,
    covariance_ab: float,
    mean_size: int,  # Same size for both groups, TODO to change?
    transformation: str,
    alternative: str = "two-sided",
) -> float:
    """
    Calculate pvalue for continuous transformation using Delta method.

    Applying the Delta Method in Metric Analytics: A Practical Guide with Novel Ideas.
    Alex Deng, Ulf Knoblich, Jiannan Lu. 2018

    Arguments
    ---------
    mean_a : float
        Metrics mean in group A
    mean_b : np.ndarray
        Metrics mean in group B
    variance_group_b
        Metrics variance in group B
    mean_size: int
        Mean size of two groups
    transformation : str
        Continuous transformation of random variable
    alternative : str, default: ``two-sided``
        Alternative for static criteria - two-sided, less, greater
        Less means, that mean in first group less, than mean in second group
    Returns
    -------
    pvalue : float
        P-value based on the first order Taylor expansion of continuous transformation
    """
    if transformation == "fraction":
        point_estimate = calculate_point_effect_by_delta_method(
            mean_a, mean_b, var_group_a, var_group_b, covariance_ab, mean_size, transformation
        )
        fraction_se: float = calc_statistic_for_delta_method(
            mean_a, mean_b, var_group_a, var_group_b, covariance_ab, mean_size
        )
        statistic: float = point_estimate / fraction_se
    else:
        raise ValueError(f"Got unknown random variable transformation: {ADMISSIBLE_TRANSFORMATIONS}")

    if alternative == "less":
        pvalue: float = sps.norm.cdf(statistic)
    elif alternative == "greater":
        pvalue: float = sps.norm.sf(statistic)
    elif alternative == "two-sided":
        pvalue: float = 2 * min(sps.norm.cdf(statistic), sps.norm.sf(statistic))
    else:
        raise ValueError(f"Incorrect alternative value - {alternative}, choose from two-sided, less, greater")
    return pvalue


def check_alternative(alternative: str) -> None:
    """
    Check correctness of alternative value
    """
    valid_alternatives: List[str] = ["two-sided", "less", "greater"]
    problem = f"Incorrect alternative value - {alternative}, choose from two-sided, less, greater"
    if alternative not in valid_alternatives:
        raise ValueError(problem)


def corrected_alpha(alpha: Union[float, np.ndarray], alternative: str) -> np.ndarray:
    """
    Corrects alpha according to alternative
    """
    if isinstance(alpha, float):
        return alpha if alternative == "two-sided" else min(2 * alpha, 1)
    else:
        return alpha if alternative == "two-sided" else np.minimum(2 * alpha, np.ones(alpha.shape[0]))


def choose_from_bounds(
    left_ci: np.ndarray,
    right_ci: np.ndarray,
    alternative: str = "two-sided",
    right_bound: float = np.inf,
    left_bound: float = -np.inf,
) -> Tuple[np.ndarray, np.ndarray]:
    """
    Choose left and right bounds according to alternative
    """
    cond_many: bool = isinstance(left_ci, Iterable)
    amount: int = len(left_ci) if cond_many else 1
    if alternative == "greater":
        right_ci = np.ones(amount) * right_bound if cond_many else right_bound
    if alternative == "less":
        left_ci = np.ones(amount) * left_bound if cond_many else left_bound
    return left_ci, right_ci


def calculate_intervals_by_delta_method(
    mean_a: float,
    mean_b: float,
    var_group_a: float,
    var_group_b: float,
    covariance_ab: float,
    mean_size: int,  # Same size for both groups, TODO to change?
    transformation: str,
    alpha: np.ndarray = np.array([0.05]),
    alternative: str = "two-sided",
) -> types.ManyIntervalType:
    """
    Computation of confidence intervals for each I type error bound (alpha) using Delta method.

    Applying the Delta Method in Metric Analytics: A Practical Guide with Novel Ideas.
    Alex Deng, Ulf Knoblich, Jiannan Lu. 2018

    Parameters
    ----------
    mean_a : float
        Metrics mean in group A
    mean_b : np.ndarray
        Metrics mean in group B
    variance_group_b
        Metrics variance in group B
    mean_size: int
        Mean size of two groups
    transformation : str
        Continuous transformation of random variable
    alpha : np.ndarray, default: ``np.array([0.05])``
        Array of I type errors bounds
    alternative : str, default ``two-sided``
        Alternative for static criteria - two-sided, less, greater
        Less means, that mean in first group less, than mean in second group

    Returns
    -------
    (left_bounds, right_bounds) : types.ManyIntervalType
        Confidence intervals based on the first order Taylor expansion
        of continuous transformation
    """
    if transformation == "fraction":
        point_estimate = calculate_point_effect_by_delta_method(
            mean_a, mean_b, var_group_a, var_group_b, covariance_ab, mean_size, transformation
        )
        fraction_se: float = calc_statistic_for_delta_method(
            mean_a, mean_b, var_group_a, var_group_b, covariance_ab, mean_size
        )
        correct_alpha: float = corrected_alpha(alpha, alternative)
        quantiles: np.ndarray = sps.norm.ppf(1 - correct_alpha / 2)
        shift: np.ndarray = quantiles * fraction_se
        left_bounds: np.ndarray = point_estimate - shift
        right_bounds: np.ndarray = point_estimate + shift
        left_bounds, right_bounds = choose_from_bounds(left_bounds, right_bounds, alternative)
    else:
        raise ValueError(f"Got unknown random variable transformation: {ADMISSIBLE_TRANSFORMATIONS}")
    conf_intervals: List[Tuple] = list(zip(left_bounds, right_bounds))
    return conf_intervals


def calculate_pvalue_by_interval(
    interval_function: Callable, criterion_value_label: float = 0, precision: float = 10e-7, **kwargs
) -> float:
    """
    Calculate pvalue for confidence interval.
    pvalue(x) = inf_a {a | x \\in S_a }
    S_a = {x | 0 not in interval(x) }.

    Parameters
    ----------
    interval_function : Callable
        Function returns confidence interval using kwargs
        Function must have argument confidence_level !!!
    criterion_value_label : float, default: ``0``
        This number indicates whether the null hypothesis should
        be rejected if it falls within the interval.
    precision : float, default: ``0.0000001``
        Precision for binary search solution.

    Returns
    -------
    pvalue : float
        P-value of the interval-induced criterion.
    """
    left: float = 0
    right: float = 1
    while right - left > precision:
        middle: float = (left + right) / 2
        conf_interval: Tuple[float, float] = interval_function(confidence_level=1 - middle, **kwargs)
        if conf_interval[0] <= criterion_value_label <= conf_interval[1]:
            left = middle
        else:
            right = middle
    return (left + right) / 2



#  Copyright 2022 MTS (Mobile Telesystems)
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""
Module contains ML-based data transformation methods for the experiment
acceleration.
"""
import json
from pathlib import Path
from typing import Any, Callable, Dict, Optional, Union

import joblib
import numpy as np
import pandas as pd
from catboost import CatBoostRegressor
from sklearn.linear_model import Ridge
from sklearn.metrics import mean_squared_error

from ambrosia import types
from ambrosia.tools import log
from ambrosia.tools.ab_abstract_component import AbstractVarianceReducer
from ambrosia.tools.back_tools import wrap_cols


class MLVarianceReducer(AbstractVarianceReducer):
    """
    Machine Learning approach for variance reduction.

    Building a model M, we can make a transformation:
    Y_hat = Y - M(X) + MEAN(M(X))

    It is important, that that the mean of M(X) do not change over time!!!
    You can choose models from Gradient boosting or Ridge regression or your
    own model class, for example ``sklearn.ensemble.RandomForest``, and pass
    models params to constructor function for a model assembly.

    Parameters
    ----------
    model : str or model type, default: ``"boosting"``
        Model which will be used for the transformations.
    model_params : Dict, optional
        Dictionary with parameters which will be used in constructor
        for a model assembly.
    scores : Dict[str, Callable], optional
        Scores which will be used.
    verbose : bool, default: ``True``
        If ``True`` will print in sys.stdout the information
        about the reduction in variance.

    Attributes
    ----------
    model : model type
        Model which will be used for the transformations.
    params : Dict
        Parameters of instance that will be updated after calling fit() method.
        Include:
        - target column name
        - covariate columns names
        - name of column after the transformation
        - additional train bias equals mean(M(X)).
    scores : Dict[str, Callable]
        Scores which will be used.
    verbose : bool
        Verbose info flag.
    fitted : bool
        Fit status flag.

    Examples
    --------
    We have data table with column 'target' and columns 'feature_1',
    'feature_2', 'feature_3'. Let us assume, that means of all these metrics
    don't change over the time, it can be age for example. We want to reduce
    variance using the predictions some of ML model, then we can use this class:

    >>> transformer = MLVarianceReducer() # By default CatBoost model will be choosen
    >>> transformer.fit_transform(dataframe, 'target', [feature columns], inplace=True, name='new_target')
    >>> transformer.store_params('path_ml_params.json')

    Now to transform the experimental data we use the following commands:

    >>> transformer = MLVarianceReducer()
    >>> transformer.load_params('path_ml_params.json')
    >>> transformer.transform(exp_data, inplace=True)

    Methods
    -------
    get_params_dict()
        Returns dict with instance fitted parameters.
    load_params_dict()
        Load parameters from the dict.
    store_params(store_path)
        Store fitted params in a json file and pickle model file.
    load_params(load_path)
        Load params from a json file and pickled model.
    fit(**fit_params)
        Fit model using a train data.
    transform(dataframe, inplace)
        Transform target column of a data frame.
    fit_transform(dataframe, **fit_params, inplace)
        Combination of fit() and transform() methods.
    """

    def __set_scorer(self, scores: Optional[Dict[str, Callable]]):
        """
        Support method for scorer setting.
        """
        if scores is not None:
            self.score = scores
        else:
            self.score = {"MSE": mean_squared_error}

    def __create_model(self) -> None:
        """
        Construct variance reducing ML model.
        """
        if not isinstance(self.model, str):
            self.model = self.model(**self.model_params)
        if self.model == "linear":
            self.model = Ridge(**self.model_params)
        if self.model == "boosting":
            if "verbose" not in self.model_params:
                self.model_params["verbose"] = False
            self.model = CatBoostRegressor(**self.model_params)

    def __init__(
        self,
        model: Union[str, Any] = "boosting",
        model_params: Optional[Dict] = None,
        scores: Optional[Dict[str, Callable]] = None,
        verbose: bool = True,
    ) -> None:
        super().__init__(verbose)
        self.params["covariate_columns"] = None
        self.params["train_bias"] = None
        self.model = model
        self.model_params = {} if model_params is None else model_params
        self.__set_scorer(scores)

    def __str__(self) -> str:
        return f"ML approach reduce for {self.params['target_column']}"

    def __call__(self, y: np.ndarray, X: np.ndarray) -> np.ndarray:
        """
        Transform target values using its predictions based on covariates.

        Class must be fitted.
        """
        self._check_fitted()
        y_hat = y - self.model.predict(X) + self.params["train_bias"]
        return y_hat

    def _verbose_score(self, dataframe: pd.DataFrame, prediction: np.ndarray) -> None:
        for name, scorer in self.score.items():
            current_score: float = scorer(dataframe[self.params["target_column"]], prediction)
            log.info_log(f"Prediction {name} score - {current_score:.5f}")

    def _check_load_params(self, params: Dict) -> None:
        for parameter in self.params:
            if parameter in params:
                self.params[parameter] = params[parameter]
            else:
                raise TypeError(f"params argument must contain: {parameter}")

    def get_params_dict(self) -> Dict:
        """
        Returns a dictionary with params.

        Returns
        -------
        params : Dict
            Dictionary with fitted params.
        """
        self._check_fitted()
        return {
            "target_column": self.params["target_column"],
            "covariate_columns": self.params["covariate_columns"],
            "transformed_name": self.params["transformed_name"],
            "train_bias": self.params["train_bias"],
            "model": self.model,
        }

    def load_params_dict(self, params: Dict) -> None:
        """
        Load instance parameters from the dictionary.

        Parameters
        ----------
        params : Dict
            Dictionary with params.
        """
        self._check_load_params(params)
        if "model" in params:
            self.model = params["model"]
        else:
            raise TypeError(f"params argument must contain: {'model'}")
        self.fitted = True

    def store_params(self, config_store_path: Path, model_store_path: Path) -> None:
        """
        Store params of model as a json file, available only for CatBoost
        model.

        You can reach model using instance.model and store it by yourself.

        Parameters
        ----------
         store_path : Path
            Path where models parameters will be stored in a json format.
        """
        self._check_fitted()
        with open(config_store_path, "w+") as file:
            json.dump(self.params, file)
        joblib.dump(self.model, model_store_path)

    def load_params(self, config_load_path: Path, model_load_path: Path) -> None:
        """
        Load models params from a json file, works only for CatBoost model.

        Parameters
        ----------
        load_path: Path
            Path to a json file with model parameters.
        """
        with open(config_load_path, "r+") as file:
            params = json.load(file)
            self._check_load_params(params)
        self.model = joblib.load(model_load_path)
        self.fitted = True

    def fit(
        self,
        dataframe: pd.DataFrame,
        target_column: types.ColumnNameType,
        covariate_columns: types.ColumnNamesType,
        transformed_name: Optional[types.ColumnNamesType] = None,
    ) -> None:
        """
        Fit model for transformations.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for model fitting.
        target_column : ColumnNameType
            Column from the dataframe, for which transformation will be
            applied.
        covariate_columns: ColumnNamesType
            Columns which will be used for the transformation.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        """
        covariate_columns = wrap_cols(covariate_columns)
        self._check_cols(dataframe, [target_column] + covariate_columns)
        self.__create_model()
        self.model.fit(dataframe[covariate_columns].values, dataframe[target_column].values)

        self.params["target_column"] = target_column
        self.params["transformed_name"] = transformed_name
        self.params["covariate_columns"] = covariate_columns
        self.params["train_bias"] = np.mean(self.model.predict(dataframe[covariate_columns].values))
        self.fitted = True

    def transform(
        self,
        dataframe: pd.DataFrame,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Transform data using the fitted model.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for transformation.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self._check_cols(dataframe, [self.params["target_column"]] + self.params["covariate_columns"])
        self._check_fitted()
        prediction: np.ndarray = self(
            dataframe[self.params["target_column"]].values, dataframe[self.params["covariate_columns"]].values
        )
        new_target: np.ndarray = prediction + np.mean(dataframe[self.params["target_column"]]) - np.mean(prediction)
        if self.verbose:
            old_variance: float = np.var(dataframe[self.params["target_column"]].values)
            new_variance: float = np.var(prediction)
            self._verbose(old_variance, new_variance)
            self._verbose_score(dataframe, prediction)
        return self._return_result(dataframe, new_target, inplace)

    def fit_transform(
        self,
        dataframe: pd.DataFrame,
        target_column: types.ColumnNameType,
        covariate_columns: types.ColumnNamesType,
        transformed_name: Optional[types.ColumnNamesType] = None,
        inplace: bool = False,
    ) -> Union[pd.DataFrame, None]:
        """
        Combinate consequentially ``fit()`` and ``transform()`` methods.

        Parameters
        ----------
        dataframe : pd.DataFrame
            Table with data for model fitting and further transformation.
        target_column : ColumnNameType
            Column from the dataframe, for which transformation will be
            applied.
        covariate_columns: ColumnNamesType
            Columns which will be used for the transformation.
        transformed_name : ColumnNamesType, optional
            Name for the new transformed target column, if is not defined
            it will be generated automatically.
        inplace : bool, default: ``False``
            If is ``True``, then method returns ``None`` and
            sets a new column for the original dataframe.
            Otherwise return copied dataframe with a new column.
        """
        self.fit(dataframe, target_column, covariate_columns, transformed_name)
        return self.transform(dataframe, inplace)
